import threading
import time

from http.server import BaseHTTPRequestHandler, HTTPServer
from multiprocessing import Value
from ctypes import c_wchar_p



class AuthorizationApp:
    """A temporary HTTP server to host a page for the redirect URI from Twitch,
    and on that, send a POST request to another internally hosted route
    with the requested user access token which would otherwise only be accessible
    from the web browser and the user
    """

    def __init__(self, host: str = 'localhost', port: int = 6319):
        self.host = host
        self.port = port

        self.server = None

        # Start the authorization process
        self.start(self.host, self.port)

    def start(self, host: str, port: int) -> None:
        """Start HTTP server to begin the authorization process"""

        TemporaryServer.status = True
        self.server = HTTPServer((self.host, self.port), TemporaryServer)

        self.thread = threading.Thread(target=self.server.serve_forever)
        self.thread.start()

    def wait_for_shutdown(self) -> None:
        """Wait for a cue to shut down server"""

        while TemporaryServer.status:
            time.sleep(0.02)

        self.server.shutdown()
        self.thread.join()

    def get_token(self) -> str:
        """Get the user access token"""
        
        # Get the token
        token = TemporaryServer.token
        # Reset token field
        TemporaryServer.token = ""

        return token


class TemporaryServer(BaseHTTPRequestHandler):
    status = Value('b', False)
    token = Value(c_wchar_p, "")

    def _set_headers(self):
        """Set Content-Type of headers"""
        self.send_response(200)
        self.send_header('Content-Type', 'text/html')
        self.end_headers()

    def do_GET(self):
        """Redirect the user to send the only locally accessible user access token
        in a POST request to another route which will receive and extract it
        """

        self._set_headers()
        self.wfile.write(b"""
            <!DOCTYPE html>
                <html lang="en">
                    <head>
                        <meta charset="UTF-8">
                        <title>mytwitch Authentication</title>
                        
                        <style>
                            body {
                                background-color: #9146FF;

                                color: white;
                                font-family: 'Ubuntu', sans-serif;
                                font-size: 2em;
                            }
                            
                            .centered {
                                margin: 0;

                                position: absolute;
                                top: 50%;
                                left: 50%;
                                transform: translate(-50%, -50%);

                                text-align: center;
                            }

                            h1, h4 { margin: 1rem; }
                            h1 { font-size: 4em; } 

                            a[href] {
                                color: white;
                                text-decoration: none;
                            }
                        </style>
                    </head>

                    <body>
                        <span class="centered">
                            <h1><a href="https://gitlab.com/thedisruptproject/mytwitch">mytwitch</a></h1>

                            <noscript>
                                <h4>
                                    Please enable JavaScript! I need to redirect
                                    the user&nbsp;access&nbsp;token back to the application.
                                </h4>
                            </noscript>
                        </span>

                        <script lang="javascript">
                            let xmlhttp = new XMLHttpRequest()
                            
                            xmlhttp.open('POST', '/', true)

                            xmlhttp.onreadystatechange = () => {
                                window.close()
                            }

                            xmlhttp.setRequestHeader('Content-Type', 'application/json')
                            xmlhttp.send(document.location.hash)
                        </script>
                    </body>
                </html>
        """)

    def do_POST(self):
        """Receive and extract the user access token
        from the pound sign section of the redirect URL
        """

        # Get body from request
        self._set_headers()
        content_len = int(self.headers.get('content-length', 0))
        body = self.rfile.read(content_len).decode(encoding='UTF-8')

        # Remove # from the start
        body = body[1:]

        # Read parameters from string
        # by splitting at &s and =s

        params = {}

        for segment in body.split('&'):
            try:
                key, value = segment.split('=')
                params[key] = value
            
            except ValueError:
                # Ignore if segment could not be parsed
                # into exactly two strings, key and value
                continue

        if 'access_token' not in params:
            raise KeyError('`access_token` could not be found')

        # Save the received user access token
        type(self).token = params['access_token']

        # Send cue that server may be shut down
        type(self).status = False

    def log_message(self, format, *args):
        # Do not log anything
        return


from typing import Union

from collections import namedtuple



_IRCMessage = namedtuple('IRCMessage', [
    'raw',
    'prefix',
    'user',
    'channel',
    'text',
    'irc_command',
    'irc_args'
])


class IRCMessage:
    """Twitch IRC message in an understandable format"""

    def __new__(cls, string: str) -> _IRCMessage:
        """Parses a string received from the Twitch IRC"""

        parts = string.split(" ")

        # Initialize all fields as None,
        # so that the message is always valid
        prefix = None
        user = None
        channel = None
        text = None
        irc_command = None
        irc_args = None


        # If the first part starts with a colon,
        # we know that it's the prefix and can set it
        if parts[0].startswith(':'):
            prefix = parts[0][1:]

            # Extract user from prefix
            user = cls.get_user_from_prefix(prefix)

            # Remove prefix from parts
            parts = parts[1:]


        # Find and extract message content
        text_start = next((i for i, part in enumerate(parts) if part.startswith(':')), None)

        if text_start is not None:
            # All words of the message text
            # lie in front of and includes this index
            text_parts = parts[text_start:]

            # Remove starting colon from first word
            text_parts[0] = text_parts[0][1:]

            # Join all words back together for the full content
            text = " ".join(text_parts)

            # Remove all text from parts
            parts = parts[:text_start]


        # Get the IRC command and arguments
        # from the remaining parts
        irc_command = parts[0]
        irc_args = parts[1:]


        # Find channel in IRC commands
        hash_start = next((i for i, part in enumerate(irc_args) if part.startswith('#')), None)

        if hash_start is not None:
            channel = irc_args[hash_start][1:]


        return _IRCMessage(
            raw=string,
            prefix=prefix,
            user=user,
            channel=channel,
            text=text,
            irc_command=irc_command,
            irc_args=irc_args
        )

    @classmethod
    def get_user_from_prefix(cls, prefix: str) -> Union[str, None]:
        """Extract user from message prefix"""
        
        domain = prefix.split('!')[0]

        if domain.endswith('.tmi.twitch.tv'):
            return domain[:-len('.tmi.twitch.tv')]

        if 'tmi.twitch.tv' not in domain:
            return domain

        return None


import argparse
from argparse import ArgumentParser

from rich import print



def printerr(text: str):
    """Print error message and exit program"""
    print(text)
    exit()


def main():
    parser = ArgumentParser(description='Terminal operations for Mytwitch')

    commands = parser.add_subparsers(dest='command')
    auth_parser = commands.add_parser('auth', help='Authentication')


    # |-----------------
    # | AUTHENTICATION

    from . import client_id
    
    # Operations
    auth_action = auth_parser.add_mutually_exclusive_group(required=True)
    auth_action.add_argument('-N', dest='new', action='store_true', help='Create new token (requires -S)')
    auth_action.add_argument('-R', dest='revoke', action='store_true', help='Revoke token (requires -T)')

    # Value parameters
    auth_parser.add_argument(
        '-C',
        default=client_id,  # Mytwitch application client ID
        dest='client_id',
        help="Twitch application client ID"
    )
    auth_parser.add_argument('-T', dest='token')
    auth_parser.add_argument('-S', dest='scope', nargs='+')


    args = parser.parse_args()

    if args.command == 'auth':

        # Create new token
        if args.new:
            if not args.scope:
                printerr(
                    '[red]Scope is required.[/red] Specify with: [b]-S[/b]\n'
                    "\tEx. [b]-N -S 'chat:read' 'chat:edit'[/b]")

            from .auth import UserToken

            print(f'{UserToken(args.client_id, args.scope)!s}')


        # Revoke token
        if args.revoke:
            if not args.token:
                printerr(
                    '[red]Token is required.[/red] Specify with: [b]-T[/b]\n'
                    "\tEx. [b]-R -T 'abcdefghijklmnopqrstuvwxyz0123456789'[/b]")

            from .auth import UserToken
            
            user_token = UserToken(args.client_id, token=args.token, immed_auth=False)
            success = user_token.revoke()

            # Output the success in lowercase
            # for other terminal applications
            # to be able to use and easily understand
            print(str(success).lower())


if __name__ == '__main__':
    main()


import webbrowser
import time
import subprocess

from typing import Union, Sequence
from urllib.parse import urlparse, quote

from .authapp import AuthorizationApp

from ..request import APIRequest

from ..exceptions import APIRequestError
from ..exceptions import InvalidToken



class UserToken:
    """A user access token for Twitch"""

    def __init__(self,
        client_id: str, scope: Union[str, Sequence] = "",
        host: str = 'localhost', port: int = 6319, token: str = None,
        immed_auth: bool = True
    ):
        self.client_id = client_id
        self.scope = scope.split(" ") if isinstance(scope, str) else scope
        
        self.host = host
        self.port = port
        
        self.token = token
        self.expires_at = None


        # All parameters required for the token request
        self.query = {
            'client_id': client_id,
            'redirect_uri': get_redirect_uri(self.host, self.port),
            'response_type': 'token',
            'scope': quote(" ").join(self.scope)
        }

        # Make sure to start with a valid token
        if immed_auth:
            str(self)

    def __str__(self) -> str:
        """Make sure the token is valid and then return it"""

        if not self.is_valid():
            self.request_new()

        return self.token

    def is_valid(self) -> bool:
        """Validate token with the Twitch API"""
        
        if not self.token:
            return False

        if self.expires_at is not None:
            return time.time() < self.expires_at

        try:
            response = APIRequest('GET',
                'https://id.twitch.tv/oauth2/validate',
                headers={'Authorization': f'Bearer {self.token}'}
            ).send()
            
            self.expires_at = time.time() + response['expires_in']
            
            # Keep username and user ID for external use
            self.login = response.get('login')
            self.user_id = response.get('user_id')

            return True

        except APIRequestError:
            return False

    def request_new(self) -> None:
        """Request a new user access token,
        and retrieve it through a temporary HTTP server
        """

        # Start an AuthorizationApp server
        auth_app = AuthorizationApp(self.host, self.port)

        # Create URL from parameters
        url = get_authorization_url(**self.query)

        # Open address in web browser
        subprocess.run(['python', '-m', 'webbrowser', '-n', url], capture_output=True)

        # Wait for the POST to be submitted
        # and the HTTP server to be closed
        auth_app.wait_for_shutdown()
        self.token = auth_app.get_token()

        # If the newly requested token is not valid, something is wrong
        if not self.is_valid():
            raise InvalidToken(
                'Something has gone wrong with retrieving a new valid token')

    def revoke(self) -> bool:
        """Revoke token"""
        response = APIRequest('POST',
            'https://id.twitch.tv/oauth2/revoke',
            data={'client_id': self.client_id, 'token': self.token}
        ).send()

        # Safety check to make sure token has been invalidated
        return not self.is_valid()



def get_authorization_url(**query) -> str:
    """Create a URL to the authentication endpoint from query"""

    # Join all parameters by &s and append to endpoint
    params = '&'.join(f'{key}={value}' for key, value in query.items())
    url = f'https://id.twitch.tv/oauth2/authorize?{params}'

    return url


def get_redirect_uri(host: str, port: int) -> str:
    """Create a URL for the redirect URI"""
    return f'http://{host}:{port}'


class TwitchAPIException(Exception):
    """Base class for errors relating to the Twitch API"""


class APIRequestError(TwitchAPIException):
    """Error included in a response from the Twitch API"""

    def __init__(self, status, message):
        super().__init__(f'[{status}] {message}')

        self.status = status
        self.message = message


class InvalidToken(TwitchAPIException):
    """User access token used for request is invalid"""


import socket
import ssl

from typing import Sequence, Union, Generator

from queue import Queue
from queue import Empty as QueueIsEmpty

from .message import IRCMessage
from ..auth import UserToken



class TwitchIRC:
    """Socket connection to the Twitch IRC"""

    def __init__(self, user_token: UserToken, channels: Sequence[str], log_commands: bool = True):
        self.user_token = user_token
        self.channels = channels
        self.log_commands = log_commands

        self.history = Queue()

    def is_connected(self) -> bool:
        """Get connection status of socket"""
        return hasattr(self, 'irc')

    def ensure_connection(self) -> None:
        """Ensure that connection is established"""
        if not self.is_connected():
            self.connect()

    def send_command(self, command: str) -> None:
        """Send a command to the IRC"""
        
        # Only print to console if not the password
        if not command.startswith('PASS'):
            if self.log_commands:
                print(f'< {command}')

        # Send command to IRC
        self.irc.send((command + '\r\n').encode())

    def send_message(self, channel: str, text: str) -> None:
        """Send a message to a channel"""
        self.send_command(f'PRIVMSG #{channel} : {text}')

    def connect(self) -> None:
        """Connect to the Twitch IRC over a socket"""

        # Connect to IRC
        self.irc = ssl.wrap_socket(socket.socket())
        self.irc.connect(('irc.chat.twitch.tv', 6697))

        # Authenticate to Twitch
        self.send_command(f'PASS oauth:{self.user_token}')
        self.send_command(f'NICK {self.user_token.login}')

        # Connect to all channels
        for channel in self.channels:
            self.send_command(f'JOIN #{channel}')

    def get_message(self) -> Union[IRCMessage, None]:
        """Get one message from the history"""

        # Refill history
        if self.history.empty():
            list(map(self.history.put, self._get_new_messages()))

        try:
            return self.history.get_nowait()
        except QueueIsEmpty:
            return None

    def _get_new_messages(self) -> Generator[IRCMessage, None, None]:
        """Get new messages"""

        messages = self.irc.recv(2048).decode()

        for raw_message in messages.split('\r\n'):
            # Ignore if empty
            if len(raw_message) == 0:
                continue

            # Parse message
            message = IRCMessage(raw_message)

            # PONG if Twitch PINGs the client
            # to make sure the connection is maintained
            if message.irc_command == 'PING':
                self.send_command('PONG :tmi.twitch.tv')

            yield message

    def feed(self):
        """Receive messages from channels and loop over them"""

        # Connect if not already connect
        self.ensure_connection()

        # Receive messages
        while True:
            for message in self._get_new_messages():
                yield message


# -------------------------------------
# -- Thanks to Willow for helping me --
# -- with asyncio and websockets     --
# --                                 --
# -- Willow                          --
# --   * Discord: Willow#1152        --
# --   * Website: aricodes.net       --
# -------------------------------------


import asyncio
import time
import json
import lzon

import websockets
from websockets import WebSocketClientProtocol
from websockets import ConnectionClosed
from asyncio.exceptions import TimeoutError as AsyncTimeoutError


from .auth import UserToken
from .exceptions import APIRequestError


class TwitchPubSub:
    """A template class for a Twitch PubSub websocket"""

    uri = 'wss://pubsub-edge.twitch.tv'

    user_token: UserToken
    topics: list[str]
    auto_reconnect: bool
    websocket: WebSocketClientProtocol

    def __init__(self, user_token: UserToken, topics: list[str],
            auto_reconnect: bool = True, timeout: float = None):
        self.user_token = user_token
        self.topics = topics
        
        self.auto_reconnect = auto_reconnect
        self.timeout = timeout
        self.websocket = None


    async def ping(self, websocket: WebSocketClientProtocol):
        """Ping websocket every 4 minutes"""
        while websocket.open:
            await websocket.send('{"type": "PING"}')
            await asyncio.sleep(240)

    async def subscribe(self, websocket: WebSocketClientProtocol, topics: list[str]):
        """Subscribe to topics"""

        await websocket.send(json.dumps({
            'type': 'LISTEN',
            'data': {
                'topics': topics,
                'auth_token': str(self.user_token)
            }
        }))

        # Await response to make sure subscription was successful
        response = json.loads(await websocket.recv())
        
        if response.get('error') != "":
            raise APIRequestError(None, response['error'])


    async def on_open(self, websocket: WebSocketClientProtocol):
        """What to do when a connection is established"""

    async def on_message(self, websocket: WebSocketClientProtocol, message: list | dict):
        """What to do when a new message is received"""

    async def on_close(self):
        """What to do when the websocket connection has been shut down"""

    async def on_error(self, websocket: WebSocketClientProtocol, exception: Exception):
        """What to do when an error occurs in the main loop"""
        raise exception


    async def __aenter__(self):
        """Establish connection and subscribe to topics."""
        
        self.websocket = await websockets.connect(self.uri)
        self.reconnect = self.auto_reconnect

        # Subscribe to all selected topics on start-up
        await self.subscribe(self.websocket, self.topics)

        # Do whatever is defined to be done
        # once a connection is established
        await self.on_open(self.websocket)

        # Start pinging in the background
        asyncio.create_task(self.ping(self.websocket))

    async def __aexit__(self, exc_type, exc_value, exc_traceback):
        """Safely shut down connection."""

        # Remove old connection
        self.websocket = None
        
        # Do whatever is defined to be done
        # when the connection has been closed
        await self.on_close()

    async def _get_message(self, websocket: WebSocketClientProtocol):
        """Get newest message. Ping or reconnect if needed."""
        
        try:
            try:
                message = await asyncio.wait_for(websocket.recv(), self.timeout)
            except AsyncTimeoutError:
                return

            response = lzon.loads(message)

            # Reconnect if necessary
            if response.get('type') == 'RECONNECT':
                self.reconnect = True
                await self.disconnect(websocket)

            # Ignore all pongs
            elif response.get('type') == 'PONG':
                return

            else:
                # Do whatever is defined to be done
                # with a message when received
                return response
            
        except websockets.ConnectionClosed as e:
            # Reraise error if it shouldn't reconnect
            # This will trigger the closing of the context
            if not self.reconnect:
                raise e

            # Reconnect by rerunning the entering of the context
            await self.__aenter__()

        except Exception as e:
            await self.on_error(websocket, e)

    async def get_message(self, websocket: WebSocketClientProtocol):
        """Wait for and return the newest regular message."""
        while True:
            message = await self._get_message(websocket)
            if message is not None:
                return message

    async def loop(self):
        """Main loop for the websocket connection"""
        
        async with self as pubsub:
            while True:
                response = await self.get_message(self.websocket)
                await self.on_message(self.websocket, response)

    async def disconnect(self, websocket: WebSocketClientProtocol):
        """Disconnect from websocket
        Must be raised within the loop
        """
        await websocket.close()
        raise ConnectionClosed(200, 'Voluntary disconnect')

    def connect(self):
        """Start the main loop and connect to the websocket"""
        asyncio.run(self.loop())


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mytwitch', 'mytwitch.auth', 'mytwitch.irc']

package_data = \
{'': ['*']}

install_requires = \
['lzon>=0.1.0,<0.2.0', 'websockets>=10.0,<11.0']

setup_kwargs = {
    'name': 'mytwitch',
    'version': '0.10.1',
    'description': 'Twitch API interaction',
    'long_description': "# mytwitch\n\n## Overview\n\n- [Security Warning](#security-warning)\n- [Installation](#installation)\n- [Application setup for authentication](#application-setup-for-authentication)\n- [Examples](#examples)\n- [Terminal Commands](#terminal-commands)\n\n\n## Security Warning\n\nAs this package allows you to easily generate a user access token with the specified permissions, it too has access to send your token off to a third party without your knowledge. You are advised to always check the source code for such operations of any application you use and make sure this isn't the case.\n\n### Where to look?\n\nIt would be best for you to search through the entire package to assure you aren't being mislead.\nHowever, this is time consuming, so if you're not inclined, here are links to the supposedly relevant files. You can also browse the files after installation to make sure the published content doesn't differ from that in the repository.\n\n- [mytwitch.auth.user_token:UserToken](https://gitlab.com/thedisruptproject/mytwitch/-/blob/main/mytwitch/auth/user_token.py)\n- [mytwitch.auth.auth_app:AuthenticationApp](https://gitlab.com/thedisruptproject/mytwitch/-/blob/main/mytwitch/auth/authapp.py)\n\n\n## Installation\n\n```sh\npython -m pip install mytwitch\n```\n\n\n## Application setup for authentication\n\n### Redirect URI\n\n1. Log in to [Twitch Developers](https://dev.twitch.tv/).\n2. Go to the **Applications** tab.\n3. Register your application with the URL as `http://localhost:` followed by the port you choose to use. By default, this package uses `6319`. This would be `http://localhost:6319`.\n\n### Client ID\n\n1. Log in to [Twitch Developers](https://dev.twitch.tv/).\n2. Go to the **Applications** tab.\n3. Select **Manage** by the application you're using.\n4. The client ID should be found on this page.\n\n**NOTE**: Client IDs are public and may be shared. You may use Mytwitch's client ID, but you're advised to set up your own application in case this were to ever get removed for any reason. Mytwitch's client ID can be imported from `mytwitch.client_id`.\n\n\n## Examples\n\n- [IRC](#irc)\n- [PubSub](#pubsub)\n- [User access token](#user-access-token)\n\n### IRC\n\n```py\nfrom mytwitch.auth import UserToken\nfrom mytwitch.irc import TwitchIRC\n\n\nclient_id = 'abcdefghijklmnopqrstuvwxyz0123456789'\nscope = ['chat:read']  # Permissions to read chat\n\n# Create a user access token for authentication\nuser_token = UserToken(client_id, scope)\n\nchannel = 'twitch'  # Which channel to connect to\nirc = TwitchIRC(user_token, [channel])  # Create your IRC\n\n# Read incoming messages\nfor message in irc.feed():\n    print(message)\n```\n\n### PubSub\n\n```py\nfrom mytwitch.auth import UserToken\nfrom mytwitch.pubsub import TwitchPubSub\n\n\nclient_id = 'abcdefghijklmnopqrstuvwxyz0123456789'\nscope = ['channel:read:redemptions']  # Permissions to read reward redemptions\n\n# Create a user access token for authentication\nuser_token = UserToken(client_id, scope)\n\n\n# Define a PubSub with your own events\nclass MyPubSub(TwitchPubSub):  # Inherit from `TwitchPubSub`\n    \n    async def on_open(self, websocket):\n        print('PubSub has been opened.')\n\n    async def on_message(self, ws, message):\n        print(f'Message received:\\n{message}\\n\\n')\n\n    async def on_close(self):\n        print('PubSub has been closed.')\n\n    async def on_error(self, ws, exception):\n        print(f'An error has occurred:\\n{exception}\\n\\n')\n\n\n# Topic for reading reward redemption\ntopics = [f'channel-points-channel-v1.{user_token.user_id}']\npubsub = MyPubSub(user_token, topics)  # Create your PubSub\n\n# Start the PubSub connection\npubsub.connect()\n```\n\n### User access token\n\n```py\nfrom mytwitch.auth import UserToken\n\nclient_id = 'abcdefghijklmnopqrstuvwxyz0123456789'\nscope = ['chat:read']\n\nuser_token = UserToken(\n    client_id,  # Application client ID\n    scope,      # Permissions you want\n\n    immed_auth = True  # You can set this to False if you don't want\n    # to generate a token on creation, as it opens a window in the browser\n)\n\n# Convert into a string to get the current token or generate a new one if necessary\nprint(f'My requested token is `{user_token}`')\n```\n\n\n## Terminal Commands\n\nThere are commands for authentication if you don't want to have to set up a file for such simple operations.\nThese commands use Mytwitch's client ID, by default, but you can specify your own with `-C`.\n\n### Create new token\n\n```sh\npython -m mytwitch auth -NS 'chat:read' 'chat:edit'\n```\n\n### Revoke token\n\n```sh\npython -m mytwitch auth -RT 'abcdefghijklmnopqrstuvwxyz0123456789'\n```",
    'author': 'Maximillian Strand',
    'author_email': 'maximillian.strand@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/OpenDisrupt/mytwitch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import requests
import json
import lzon

from typing import Callable

from .exceptions import APIRequestError



class APIRequest:
    """An HTTP request to the Twitch API"""

    def __init__(self, method: Callable, *args, **kwargs):
        # Use the corresponding function from requests
        if method == 'GET':
            self.method = requests.get
        elif method == 'POST':
            self.method = requests.post
        elif method == 'DELETE':
            self.method = requests.delete
        else:
            raise ValueError('Method must be either GET, POST, or DELETE')

        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.send()

    def send(self):
        """Send the request and return the response"""

        # Send request and unpack JSON payload
        response = self.method(*self.args, **self.kwargs)

        # Try to parse content if body is returned
        try:
            content = json.loads(response.content)
        except json.JSONDecodeError:
            content = {}
        
        # Twitch API error
        if 'status' in content:
            raise APIRequestError(
                content['status'],
                content['message']
            )

        return lzon.loads(content)